/* 
 * Copyright (c) 1998 The Martin Kuzela (kuzela@adela.sk).  All rights reserved. 
 * Modified by Marcelo Ochoa 1999 (mochoa@exa.unicen.edu.ar) 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 * 
 * 3. All advertising materials mentioning features or use of this 
 *    software must display the following acknowledgment: 
 *    "This product includes software developed by the OWSKiller  
 *    Project for use in the OWSkiller servlet engine project 
 *    (http://www.adela.sk/kuzela/OWSKiller) 
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)." 
 * 
 * 4. The names "OWSKiller", "OWSKiller Servlet Engine" and  
 *    "OWSKiller Project" must not be used to endorse or promote products  
 *    derived from this software without prior written permission. 
 * 
 * 5. Products derived from this software may not be called "OWSKiller" 
 *    nor "OWSKiller Server" appear in their names without  
 *    prior written permission of the OWSKiller Project. 
 * 
 * 6. Redistributions of any form whatsoever must retain the following 
 *    acknowledgment: 
 *    "This product includes software developed by the OWSKiller  
 *    Project for use in the OWSKiller servlet engine project 
 *    (http://www.adela.sk/kuzela/OWSKiller) 
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)" 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE OWSKILLER PROJECT "AS IS" AND ANY 
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR 
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * This software consists of voluntary contributions made by many 
 * individuals on behalf of the OWSKiller Group. For more information 
 * on the OWSKiller Project and the OWSKiller Servlet Engine project, 
 * please see <http://www.adela.sk/kuzela/OWSKiller/> 
 * (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller). 
 * 
 */ 
 
package ows; 
 
import java.io.*; 
import java.sql.*; 
import java.util.*; 
import java.text.DateFormat; 
import java.text.ParsePosition; 
import javax.servlet.*; 
import javax.servlet.http.*; 
 
public class OWSKiller extends HttpServlet 
{ 
  // The ResourceManager extends the ConnectionCache class adding Transaction Support 
  // For syntax of definition of transaction see OWSKiller.cfg file 
  protected ResourceManager cache; 
 
  Base64 b64; 
 
  // From Package definition OWA_SEC 
  /*******************************************************************/ 
  /* PL/SQL Agent's authorization schemes                            */ 
  /*******************************************************************/ 
  // NO_CHECK    constant integer := 1; /* no authorization check             */ 
  // GLOBAL      constant integer := 2; /* global check by a single procedure */ 
  // PER_PACKAGE constant integer := 3; /* use auth procedure in each package */ 
  static final int NO_CHECK = 1; 
  static final int GLOBAL = 2; 
  static final int PER_PACKAGE = 3; 
  static String WEBMASTER; 
  static String CHARSET; 
  static String CONTENTTYPE; 
 
  // The Context have the definition (getServerInfo method) of CGI var 
  // SERVER_SOFTWARE need by owa.get_cgi_env 
  static ServletContext sctx; 
  // Store the SCRIPT_NAME cgi var 
  // In Apache Jserv work fine in the first call of req.getServletPath 
  // but don't work in other calls, posible bugs? 
  // with servletrunner work fine in all calls 
  static String ScriptName = null; 
  // Array for converting months name from cookies to int value of Calendar objects 
  static String []MonthToInt = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"}; 
                
 
  public void init(ServletConfig sc) 
    throws javax.servlet.ServletException 
  { 
    super.init(sc); 
    // Save the context for future invocation of method getServerInfo in method doPost 
    sctx = sc.getServletContext(); 
 
    //System.out.println("initializing"); 
 
    Properties props = new Properties(); 
    try 
    { 
      //System.out.println("File open "+sc.getInitParameter("propertiesfile")); 
      FileInputStream in = new FileInputStream(sc.getInitParameter("propertiesfile")); 
      props.load(in); 
      in.close(); 
    } catch (IOException e) 
    { 
      System.out.println("Error loading OWSKiller.cfg file"); 
      e.printStackTrace(); 
    } 
 
    // Set global OWSKiller variables 
    WEBMASTER = props.getProperty("OWSKiller.webmaster"); 
    CHARSET = props.getProperty("OWSKiller.charset"); 
    CONTENTTYPE = props.getProperty("OWSKiller.contenttype"); 
 
    // Create BASE64 decoding class 
    b64 = new Base64(); 
 
    // Initialize ConnectionCache class 
    try 
    { 
         cache = ResourceManager.getTxInstance(props); 
    } 
    catch (Exception e) 
    { 
      System.out.println("Could not load driver"); 
      e.printStackTrace(); 
    } 
  } 
 
 
 
  public void destroy() 
  { 
    try 
    { 
      cache.release(); 
    } 
    catch (Exception e) 
    { 
      System.out.println("Could not destroy driver"); 
      e.printStackTrace(); 
    } 
  } 
 
 
 
  public void doGet(HttpServletRequest req, HttpServletResponse res) 
    throws ServletException, IOException 
  { 
    doPost(req, res); 
  } 
 
 
 
  public String replace2(String s) 
  { 
    int i = 0; 
 
    //while ((i = s.indexOf(''',i)) != -1) // doesn't work under JDK 1.2 
    //while ((i = s.indexOf('\u0027',i)) != -1) // doesn't work under Oracle JDeveloper 2.0 
    while ((i = s.indexOf('\'',i)) != -1)    // "'" char 
    { 
      s = s.substring(0,i + 1) + s.substring(i); 
      i+=2; 
    } 
    return s; 
  } 
 
  // Process cookie definicion from owa_cookie.send or owa_cookie.remove 
  // Syntax definition from http://www.netscape.com/newsref/std/cookie_spec.html 
  // See owa_cookie package 
  // Don't work with cookie definition greater than 255 char's (muti-line cookies) 
  private Cookie Make_Cookie(String s) 
  { Cookie choc_chip; 
    long age; 
    // Divide the string cookie format in fields by the ; 
    StringTokenizer st = new StringTokenizer(s,";"); 
    // the name = value pairs is required 
    String s1 = (String) st.nextElement(); 
    choc_chip = new Cookie(s1.substring(0,s1.indexOf("=")),s1.substring(s1.indexOf("=")+1)); 
    //System.out.println("Name =>" + choc_chip.getName()); 
    //System.out.println("Value =>" + choc_chip.getValue()); 
    while(st.hasMoreElements()) { 
      s1 = (String) st.nextElement(); 
      // Proccess the expires field 
      if (s1.startsWith(" expires=")) { 
         s1 = s1.substring(s1.indexOf(",")+2); 
         try { 
            // Convert the Date especification to Age format of Servlets Cookie 
            // Acording to non deprected api of JDK 1.1 
            Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT")); 
            String month = s1.substring(3,6); 
            int monthint; 
       for (monthint=11;monthint>0 && !MonthToInt[monthint].equals(month);monthint--) ; 
            cal.set(new Integer(s1.substring(7,9)).intValue()+1900, 
                    monthint, 
                    new Integer(s1.substring(0,2)).intValue(), 
                    new Integer(s1.substring(10,12)).intValue(), 
                    new Integer(s1.substring(13,15)).intValue(), 
                    new Integer(s1.substring(16,18)).intValue()); 
            java.util.Date expire = cal.getTime(); 
            if (cal.get(1)==1990 && cal.get(2)==0 && cal.get(0)==1) { 
                // Option remove cookie 
                age = 0; 
            } else { 
                // Session cookie or expire in the future 
                java.util.Date now = new java.util.Date(); 
                age = (expire.getTime()-now.getTime())/1000; 
                age = (age < 0 ? -1 : age); 
            } 
            choc_chip.setMaxAge((int)age); 
            //System.out.println("Age =>" + choc_chip.getMaxAge()); 
         } catch(Exception e) { 
            System.out.println("Invalid Date format en cookie String"); 
         } 
      } // end if expires= 
      // Proccess the path field 
      if (s1.startsWith(" path=")) { 
         // Set Path 
         choc_chip.setPath(s1.substring(s1.indexOf("=")+1)); 
         //System.out.println("Path =>" + choc_chip.getPath()); 
      } // end if path= 
      // Proccess the domain field 
      if (s1.startsWith(" domain=")) { 
         // Set Domain 
         choc_chip.setDomain(s1.substring(s1.indexOf("=")+1)); 
         //System.out.println("Domain =>" + choc_chip.getDomain()); 
      } // end if domain= 
      // Proccess the secure flags 
      if (s1.startsWith(" secure")) { 
         // Set Secure 
         choc_chip.setSecure(true); 
         //System.out.println("Secure"); 
      } // end if secure 
    } // end while st.hasMoreElements 
    // Return the cookie to the caller 
    return choc_chip; 
  } 
 
  public void doPost(HttpServletRequest req, HttpServletResponse res) 
    throws ServletException, IOException 
  { 
    Connection conn = null; 
    PrintWriter out = null; 
    String command = null; 
 
    String servletpath = replace2(req.getPathInfo()); 
    //System.out.println("servletpath = " + servletpath); 
 
    String servletname = servletpath.substring(servletpath.lastIndexOf('/') + 1); 
    //System.out.println("servletname = " + servletname); 
    //System.out.println("servletpath = " + replace2(req.getServletPath())); 
 
    String ppackage; 
    String pprocedure; 
    int i = servletname.indexOf('.'); 
    // Handle anonymous Procedure 
    if (i<0) { 
       ppackage = ""; 
       pprocedure = servletname; 
    } else { 
      ppackage = servletname.substring(0, i); 
      pprocedure = servletname.substring(i + 1); 
    } 
 
    //System.out.println("pkg = " + ppackage + " proc= " + pprocedure); 
 
    // Get the writer 
    out = res.getWriter(); 
    if (ScriptName==null) { 
        ScriptName = req.getServletPath(); 
    } 
 
    String str; 
    try { 
        str = b64.decode(req.getHeader("Authorization").substring(6)); 
    } catch (Exception e) { 
        str = ":"; 
    } 
    i = str.indexOf(':'); 
    String name = str.substring(0,i); 
    String password = str.substring(i + 1); 
 
    try 
    { 
      if (!cache.NeedLogin()) // Tradicional behavior 
         conn = cache.get(req); 
      else {  // get user password at logon 
        if (name.equals("") || password.equals("")) { 
                // Require authorization 
                sendUnauthorized(res,"new DB Login"); 
                return; 
        } else { 
                // get one connection 
                conn = cache.get(req,name,password); 
        } 
      } 

      CallableStatement cs; 
      if (!ResourceManager.isTxEnable(conn)) { 
        // if the connection is normal (not inside the transaction) force to reset packages states 
        command =  "BEGIN dbms_session.reset_package; END;"; 
        cs = conn.prepareCall(command); 
        cs.execute(); 
        cs.close(); //don't wait for garbage collector 
      } 
 
      // Declare temporary vars for passing arguments to owa.init_cgi_env procedure 
      command =  "DECLARE var_val owa.vc_arr;"; 
      command += "var_name owa.vc_arr;"; 
      command += "dummy_num_vals integer; \nBEGIN "; 
      // Get dummy val, force to execute init code of the package 
      // if not execute this call the global vars of packages of owa_init 
      // and owa_cookie have null vals 
      command += "dummy_num_vals := owa.initialize;\n"; 
 
      String hostaddr = req.getRemoteAddr(); 
 
      //System.out.println(hostaddr); 
 
      //!!! make faster right into "command" variable 
      StringTokenizer st = new StringTokenizer(hostaddr,"."); 
      for(i = 1; st.hasMoreElements(); i++) 
      { 
        command += "owa.ip_address(" + i + "):='" + (String) st.nextElement() + "';\n"; 
      } 
 
      // Set the owa.cgi_var_val and owa.cgi_var_name used by owa package 
      // for example owa.get_service_path use the CGI var SCRIPT_NAME 
      command += " owa.user_id:='" + name + 
            "'; owa.password:='" + password + 
            "'; owa.hostname:='" + req.getRemoteHost() + 
            "'; htp.rows_in:=0; htp.rows_out:=0;\n" + 
            "   var_val(1):='" + req.getMethod() + "';\n" + 
            "   var_name(1):='REQUEST_METHOD';\n" + 
            "   var_val(2):='" + req.getPathInfo() + "';\n" + 
            "   var_name(2):='PATH_INFO';\n" + 
            "   var_val(3):='" + req.getPathTranslated() + "';\n" + 
            "   var_name(3):='PATH_TRANSLATED';\n" + 
            "   var_val(4):='" + req.getQueryString() + "';\n" + 
            "   var_name(4):='QUERY_STRING';\n" + 
            "   var_val(5):='" + name + "';\n" + 
            "   var_name(5):='REMOTE_USER';\n" + 
            "   var_val(6):='" + ((name.equals("") && password.equals("")) ? req.getAuthType() : "Basic") + "';\n" + 
            "   var_name(6):='AUTH_TYPE';\n" + 
            "   var_val(7):='"+ ScriptName + "';\n" + 
            "   var_name(7):='SCRIPT_NAME';\n" + 
            "   var_val(8):='"+ sctx.getServerInfo() + "';\n" + 
            "   var_name(8):='SERVER_SOFTWARE';\n" + 
            "   var_val(9):='"+ req.getContentLength() + "';\n" + 
            "   var_name(9):='CONTENT_LENGTH';\n" + 
            "   var_val(10):='"+ req.getContentType() + "';\n" + 
            "   var_name(10):='CONTENT_TYPE';\n" + 
            "   var_val(11):='"+ req.getProtocol() + "';" + 
            "   var_name(11):='SERVER_PROTOCOL';\n" + 
            "   var_val(12):='"+ req.getServerName() + "';" + 
            "   var_name(12):='SERVER_NAME';\n" + 
            "   var_val(13):='"+ req.getServerPort() + "';" + 
            "   var_name(13):='SERVER_PORT';\n" + 
            "   var_val(14):='"+ hostaddr + "';\n" + 
            "   var_name(14):='REMOTE_ADDR';\n" + 
            "   var_val(15):='"+ req.getRemoteHost() + "';\n" + 
            "   var_name(15):='REMOTE_HOST';\n" + 
            "   var_val(16):='" + req.getHeader("User-Agent") + "';\n" + 
            "   var_name(16):='HTTP_USER_AGENT';\n";
            if (req.getHeader("Cookie")!=null) {
                    command += "   var_val(17):='" + req.getHeader("Cookie") + "';\n" + 
                               "   var_name(17):='HTTP_COOKIE';\n" +
                               "   owa.init_cgi_env(17,var_name,var_val);\n";
            } else {
                    command += "   owa.init_cgi_env(16,var_name,var_val);\n "; 
            }

      // get authorization mode 
      command += "? := my_ows.get_auth_scheme; END;"; 
      cs = conn.prepareCall(command); 
      cs.registerOutParameter(1, Types.INTEGER); 
      cs.execute(); 
      // From package specification owa_sec 
      /*******************************************************************/ 
      /* PL/SQL Agent's authorization schemes                            */ 
      /*******************************************************************/ 
      //NO_CHECK    constant integer := 1; /* no authorization check             */ 
      //GLOBAL      constant integer := 2; /* global check by a single procedure */ 
      //PER_PACKAGE constant integer := 3; /* use auth procedure in each package */ 
      // Get Atuhorizacion Scheme 
      int authMode = cs.getInt(1); 
      cs.close(); //don't wait for garbage collector 
      //System.out.println("auth="+authMode); 
      command = "BEGIN "; 
      if (authMode==NO_CHECK) { // Authorization none 
         // the function my_ows.authorize always return true 
              command += "? := my_ows.b2n(" + "my_ows" + ".authorize); END;"; 
          } else if (authMode==GLOBAL) { // Authorization global 
               // In global authentication all executes are validates by the owa_init.authorize function 
                    command += "? := my_ows.b2n(" + "owa_init" + ".authorize); END;"; 
                  } else { // Authorization per package 
                    if (ppackage.equals("")) // if anonymous procedure same as NONE 
                      command += "? := my_ows.b2n(" + "my_ows" + ".authorize); END;"; 
                    else // else call to function authorize on this package 
                      command += "? := my_ows.b2n(" + ppackage + ".authorize); END;"; 
                  } 
 
      //System.out.println("cm="+command); 
      cs = conn.prepareCall(command); 
      cs.registerOutParameter(1, Types.INTEGER); 
      //((OracleStatement)cs).defineColumnType(1, Types.INTEGER); 
      cs.execute(); 
 
      // Check if password is valid 
      int authStatus = cs.getInt(1); 
      cs.close(); //don't wait for garbage collector 
      //System.out.println("Auth Status="+authStatus); 
      if (authStatus != 1) 
      { 
         // if the autorization fail get the realm from the package owa_sec 
         // The realm is set by the use owa_sec.set_protection_realm 
         // A good place for call the owa_sec.set_protection_realm is in the function authorize 
         // Ej 
         // package body example is 
         // function authorize return boolean is 
         // begin 
         //    owa_sec.set_protection_realm('Sample App.'); 
         //    if (owa_sec.get_user_id=="marcelo") then 
         //        return true; 
         //    else 
         //        return false; 
         //    end if; 
         // end authorize; 
         // end example; 
         command = "BEGIN "; 
         command += "? := my_ows.get_protection_realm; END;"; 
         cs = conn.prepareCall(command); 
         cs.registerOutParameter(1, Types.VARCHAR); 
         //((OracleStatement)cs).defineColumnType(1, Types.VARCHAR); 
         cs.execute(); 
         // Get Protection realm 
         String Realm = cs.getString(1); 
         cs.close(); //don't wait for garbage collector 
         sendUnauthorized(res,Realm); 
         return; 
      } 
 
      // parse all FORM input parameters and arrays set as PL/SQL arrays 
 
      // Calling with constants - no prepared calls 
 
      // Handling Case Insensitive args in PL/SQL and owa_image.point 
      // Ej: 
      // http://server:port/servlet/plsql/example.print?a=b 
      // http://server:port/servlet/plsql/example.print?A=b 
      // make the same call to the procedure example.print('b') 
      // PLSQL runtime choose the correct procedure to call 
      // Work with overload procedure to. 
      // Ej: 
      // http://server:port/servlet/plsql/example.print?A=b 
      // http://server:port/servlet/plsql/example.print?A=b&c=d 
 
      // Build procedure call 
      command = servletname + "("; // Main calling command 
      String decvar = "DECLARE \n";  //we will declare array variables here 
      String setvar = "BEGIN \n";    //we will set array variables here 
 
      int foundcount = 0; 
 
      // Build procedure call parameter by parameter 
      Enumeration real_args = req.getParameterNames(); 
      while(real_args.hasMoreElements()) { 
        String name_args = (String)real_args.nextElement(); 
        String multi_vals[] = req.getParameterValues(name_args); 
        if (multi_vals.length>1) { // must be owa_util.ident_array type 
          for (i = 0;i<multi_vals.length;i++) { 
            String s = new String(multi_vals[i].getBytes("8859_1"), CHARSET); 
            s = replace2(s); 
            setvar += name_args + "(" + (i+1) + "):='" + s + "'; "; 
          } // end for make array variable 
          command += name_args + "=>" + name_args + ","; 
          decvar += name_args + " owa_util.ident_arr;\n"; 
          foundcount++; 
        } else if (name_args.indexOf(".")>0) { // must be owa_image.point 
              // handle owa_image.point data type 
              name_args = name_args.substring(0,name_args.indexOf(".")); 
              decvar += name_args + " owa_image.point;\n"; 
              String val_x = req.getParameter(name_args+".x"); 
              String val_y = req.getParameter(name_args+".y"); 
              // the owa_image.point data type is a array of varchar index by binary integer 
              // Position 1 is args.x value 
              // Position 2 is args.y value 
              String s = new String(val_x.getBytes("8859_1"), CHARSET); 
              s = replace2(s); 
              setvar += name_args + "(1):='" + s +"'; "; 
              //System.out.println("point " + name_args + ".x=" + val_x); 
              s = new String(val_y.getBytes("8859_1"), CHARSET); 
              s = replace2(s); 
              setvar += name_args + "(2):='" + s +"'; "; 
              //System.out.println("point " + name_args + ".y=" + val_y); 
              command += name_args + "=>" + name_args + ","; 
              name_args = (String)real_args.nextElement(); // Skip second args 
              foundcount++; 
        } else { // otherwise, must be scalar type 
            String s = new String(req.getParameter(name_args).getBytes("8859_1"), CHARSET); 
            s = replace2(s); 
            command += name_args + "=>'" + s + "',"; 
            foundcount++; 
            //System.out.println("single "+name_args+"="+req.getParameter(name_args)); 
        } // end if muti valued args 
      } 
 
      command = decvar + setvar + command.substring(0,command.length() - 1);       
      if (foundcount == 0) 
        command +=  "; END;"; 
      else  
         command += "); END;"; 
 
      System.out.println("command="+command); 
      // Exec procedure in DB 
      cs = conn.prepareCall(command); 
      cs.execute(); 
      cs.close(); //don't wait for garbage collector 
      // Get how much of output is in htbuf 
      cs = conn.prepareCall("BEGIN ? := my_ows.howmuch; END;"); 
      cs.registerOutParameter(1, Types.INTEGER); 
      //((OracleStatement)cs).defineColumnType(1, Types.INTEGER); 
      cs.execute(); 
      int howmuch = cs.getInt(1); 
      cs.close(); //don't wait for garbage collector 
 
      //System.out.println("howmuch="+ howmuch); 
 
      // Get and print html output 
      cs = conn.prepareCall ("BEGIN ? := htp.htbuf(?); END;"); 
      //((OracleStatement)cs).defineColumnType(1, Types.VARCHAR); 
 
 
      //Check if mime type is set 
      i = 1; 
      cs.registerOutParameter(1, Types.VARCHAR); 
      cs.setInt(2,i); 
      cs.execute(); 
      //System.out.println("Execute complete"); 
 
      String s = cs.getString(1); 
      cs.close(); 
      //System.out.println(s); 
 
      if (s.startsWith("Content-type: ")) 
      { 
        s = s.substring(14, s.indexOf('\n')); 
        //System.out.println("Content-type: "+s); 
        res.setContentType(s); 
        // out = res.getWriter(); 
   // Verify if the position 2..n have the Syntax "Set-Cookie: " 
   // in this case the line is a cookie definition made it by 
   // calling owa_cookie.send(cookiename,val,...) 
        // Handle Cookie definicion header in theory 0..howmuch-1 
        cs = conn.prepareCall ("BEGIN ? := htp.htbuf(?); END;"); 
        i = 2; 
        while (i<howmuch) { 
            cs.registerOutParameter(1, Types.VARCHAR); 
            cs.setInt(2,i++); 
            cs.execute(); 
            s = cs.getString(1); 
            //System.out.println(s); 
            if (s.startsWith("Set-Cookie: ")) { 
                   // Parse the cookie line 
                   Cookie choc_chip = Make_Cookie(s.substring(12,s.indexOf('\n'))); 
                   res.addCookie(choc_chip);
            } else break; 
        } 
        cs.close(); 
      } 
      else if (s.startsWith("Location: ")) 
      { 
        s = s.substring(10, s.indexOf('\n')); 
        res.sendRedirect(res.encodeRedirectUrl(s)); 
        //System.out.println("Location:"); 
        return; 
      } else 
      { 
        res.setContentType(CONTENTTYPE); 
        // out = res.getWriter(); 
        out.print(s); 
        i = 2; 
        //System.out.println("Other: "+s); 
      } 
 
      // Get the HTML ouput in pieces of 32768 bytes , better performance!! 
      // the function my_ows.get_buf join the buf of html ouput in pieces of 
      // 32kbytes this function is faster than java client call. 
      cs = conn.prepareCall ("BEGIN my_ows.get_buf(?,?,?); END;"); 
      while( i <= howmuch ) 
      { 
        cs.registerOutParameter(1, Types.VARCHAR); 
        cs.registerOutParameter(2, Types.INTEGER); 
        cs.setInt(3,i); 
        cs.execute(); 
        s = cs.getString(1); 
        i = cs.getInt(2); 
        //System.out.println(i); 
        //System.out.println(s.length()); 
        //System.out.println(s); 
        out.print(s); 
      } 
      cs.close(); //don't wait for garbage collector 
 
    } 
    catch (Exception e) 
    { 
      String text_error = "\n\n\n While try to execute "+servletname; 
      text_error += "\n with args\n"; 
      Enumeration real_args = req.getParameterNames(); 
      while(real_args.hasMoreElements()) { 
        String name_args = (String)real_args.nextElement(); 
        text_error += "\n" + name_args + ":"; 
        String multi_vals[] = req.getParameterValues(name_args); 
        if (multi_vals.length>1) { // must be owa_util.ident_array type 
          for (i = 0;i<multi_vals.length;i++) { 
            text_error += "\n\t" + new String(multi_vals[i].getBytes("8859_1"), CHARSET); 
          } 
        } else { 
            text_error += "\n\t" + req.getParameter(name_args); 
        } 
      } 
      System.out.println("PL/SQL Error:" + new java.util.Date()); 
      //if (command != null) 
      System.out.println(command); 
      e.printStackTrace(); 
      sendFailure(res, "PL/SQL Error <br><pre>" +e.getMessage()+text_error+"</pre><br>"); 
    } 
    finally 
    { 
      try 
      { 
      if (conn != null) {cache.release(conn,req);} 
      } 
      catch (Exception e) 
      { 
        System.out.println("Error2"); 
        e.printStackTrace(); 
      } 
    } 
 
  // System.out.println("flushing"); 
  out.flush(); 
  //out.close();   //!!! close? 
 
  } 
 
 
 
  public String getServletInfo() 
  { 
    return "OWSKiller servlet (c) Martin Kuzela 1998"; 
  } 
 
 
 
  private void sendFailure(HttpServletResponse res, String reason) 
    throws IOException 
  { 
    PrintWriter out = res.getWriter(); 
    res.setContentType(CONTENTTYPE); 
    out.println("<html><head></head><body BGCOLOR=#ffffff>"); 
    out.println(reason); 
    out.println("<br>Send email to <a href=\"mailto:" + WEBMASTER + "\">" + WEBMASTER + "</a>"); 
    out.println("<br><A HREF=\"javascript:window.location.reload()\">Reload Page</A>"); 
    out.println("<br><br><A HREF=/>Home Page</A>"); 
    out.println("</body></html>"); 
  } 
 
 
 
  public static void sendUnauthorized(HttpServletResponse res,String Realm) 
    throws IOException 
  { 
    PrintWriter out = res.getWriter(); 
 
    res.setStatus(HttpServletResponse.SC_UNAUTHORIZED); 
 
    // put your realm here 
    res.setHeader("WWW-authenticate","basic realm=\"" + Realm + "\""); 
    out.print("Type your username and password to access " + Realm + " system"); 
    out.flush(); 
  } 
} 
